import os
import sys

path = os.path.dirname(__file__)
sys.path.insert(0, path)
from PolyRound.mutable_classes.polytope import Polytope
from PolyRound.static_classes.constraint_removal_reduction import PolytopeReducer
from typing import Dict
from PolyRound.static_classes.lp_utils import ChebyshevFinder
from PolyRound.static_classes.rounding.maximum_volume_ellipsoid import (
    MaximumVolumeEllipsoidFinder,
)
from PolyRound.default_settings import (
    default_hp_flags,
    default_0_width,
    default_SVD_threshold,
)
import numpy as np
from cobra.core.model import Model
from PolyRound.static_classes.parse_sbml_stoichiometry import StoichiometryParser
from PolyRound.static_classes.hdf5_csv_io import HDF5, CSV
from cobra.util.solver import solvers

if "gurobi" in solvers:
    pref_backend = "gurobi"
else:
    pref_backend = "glpk"


class PolyRoundApi:
    @staticmethod
    def simplify_polytope(
        polytope: Polytope,
        backend: str = pref_backend,
        hp_flags: Dict = default_hp_flags,
        thresh: float = default_0_width,
        verbose: bool = False,
        sgp: bool = False,
    ) -> Polytope:
        """
        Remove redundant constraints and refunction inequality constraints to equality constraints in case of dimension
        width less than thresh
        @param polytope:
        @param backend:
        @param hp_flags:
        @param thresh:
        @param verbose:
        @param sgp:
        @return:
        """
        polytope = polytope.copy()
        polytope.normalize()
        removed, refunctioned = 1, 1
        while removed != 0 or refunctioned != 0:
            polytope, removed, refunctioned = PolytopeReducer.constraint_removal(
                polytope,
                hp_flags,
                backend=backend,
                thresh=thresh,
                verbose=verbose,
                sgp=sgp,
            )
        if polytope.A.shape[0] == 0:
            raise ValueError(
                "All inequality constraints are redundant, implying that the polytope is a single point."
            )
        return polytope

    @staticmethod
    def transform_polytope(
        polytope: Polytope,
        backend: str = pref_backend,
        hp_flags: Dict = default_hp_flags,
        verbose: bool = False,
        sgp: bool = False,
    ) -> Polytope:
        """
        Express polytope in a (shifted) orthogonal basis in the null space of the equality constraints to remove all
        equality constraints
        @param polytope:
        @param backend:
        @param hp_flags:
        @param verbose:
        @param sgp:
        @return:
        """
        if polytope.inequality_only:
            raise ValueError(
                "Polytope already transformed (only contains inequality constraints)"
            )
        polytope = polytope.copy()
        x, dist = ChebyshevFinder.chebyshev_center(
            polytope, backend=backend, regularize=False, sgp=sgp, **hp_flags
        )
        if verbose:
            print("chebyshev distance is : " + str(dist))
            pre_b_dist = polytope.border_distance(x)
            print("border distance pre-transformation is: " + str(pre_b_dist))
        # put x at zero!
        polytope.apply_shift(x)
        if verbose:
            x_0 = np.zeros(x.shape)
            b_dist_at_zero = polytope.border_distance(x_0)
            print("border distance zero-transformation is: " + str(b_dist_at_zero))
        stoichiometry = polytope.S.values
        transformation = PolytopeReducer.null_space(
            stoichiometry, eps=default_SVD_threshold
        )
        polytope.apply_transformation(transformation)
        if verbose:
            u = np.zeros((transformation.shape[1], 1))
            norm_check = np.linalg.norm(np.matmul(stoichiometry, transformation))
            print("norm of the null space is: " + str(norm_check))
            b_dist = polytope.border_distance(u)
            print("border distance after transformation is: " + str(b_dist))
            # test if we can reproduce the original x
            trans_x = polytope.back_transform(u)
            x_rec_diff = np.max(trans_x - np.squeeze(x))
            print("the deviation of the back transform is: " + str(x_rec_diff))
        return polytope

    @staticmethod
    def round_polytope(
        polytope: Polytope,
        backend: str = pref_backend,
        hp_flags: Dict = default_hp_flags,
        verbose: bool = False,
        sgp: bool = False,
    ) -> Polytope:
        """
        Round polytope using the maximum volume ellipsoid approach
        @param polytope:
        @param backend:
        @param hp_flags:
        @param verbose:
        @param sgp:
        @return:
        """
        # check if there are Nans
        bool = False
        bool += np.isinf(polytope.A.values).any()
        bool += np.isinf(polytope.b.values).any()
        if bool:
            raise ValueError("Polytope assigned for rounding contains inf")
        polytope = polytope.copy()
        MaximumVolumeEllipsoidFinder.iterative_solve(
            polytope, backend, hp_flags=hp_flags, verbose=verbose, sgp=sgp
        )
        return polytope

    @staticmethod
    def simplify_transform_and_round(
        polytope: Polytope,
        backend: str = pref_backend,
        hp_flags: Dict = default_hp_flags,
        thresh: float = default_0_width,
        verbose: bool = False,
        sgp: bool = False,
    ) -> Polytope:
        """
        Conveniently execute simplify_polytope, transform_polytope and round polytope in sequence
        @param polytope:
        @param backend:
        @param hp_flags:
        @param thresh:
        @param verbose:
        @param sgp:
        @return:
        """
        polytope = PolyRoundApi.simplify_polytope(
            polytope,
            backend=backend,
            hp_flags=hp_flags,
            thresh=thresh,
            verbose=verbose,
            sgp=sgp,
        )
        polytope = PolyRoundApi.transform_polytope(
            polytope, backend=backend, hp_flags=hp_flags, verbose=verbose, sgp=sgp
        )
        polytope = PolyRoundApi.round_polytope(
            polytope, backend=backend, verbose=verbose, sgp=sgp
        )
        return polytope

    @staticmethod
    def cobra_model_to_polytope(model: Model):
        """
        Turn cobrapy model into polytope
        @param model:
        @return:
        """
        return StoichiometryParser.extract_polytope(model)

    @staticmethod
    def polytope_to_hdf5(polytope: Polytope, filename: str):
        HDF5.polytope_to_h5(polytope, filename)

    @staticmethod
    def polytope_to_csvs(polytope: Polytope, dirname: str):
        CSV.polytope_to_csv(polytope, dirname)

    @staticmethod
    def sbml_to_polytope(file_name: str) -> Polytope:
        polytope = StoichiometryParser.parse_sbml_cobrapy(file_name)
        return polytope



import pandas as pd
import os
from PolyRound.mutable_classes.polytope import Polytope
from pathlib import Path
import h5py


class CSV:
    @staticmethod
    def polytope_to_csv(polytope, dirname):
        Path(dirname).mkdir(parents=True, exist_ok=True)
        name = dirname.rstrip("/").split("/")[-1]
        for attribute in dir(polytope):
            tentative_df = getattr(polytope, attribute)
            if isinstance(tentative_df, pd.DataFrame) or isinstance(
                tentative_df, pd.Series
            ):

                if attribute == "transformation":
                    zero_solution_df = pd.Series(0, index=tentative_df.columns)
                    zero_solution_df.to_csv(
                        os.path.join(dirname, "start_" + name + "_rounded.csv"),
                        header=False,
                        index=False,
                    )
                    tentative_df.to_csv(
                        os.path.join(dirname, "N_" + name + "_rounded.csv"),
                        header=False,
                        index=False,
                    )
                elif attribute == "shift":
                    tentative_df.to_csv(
                        os.path.join(dirname, "p_shift_" + name + "_rounded.csv"),
                        header=False,
                        index=False,
                    )
                    name_series = pd.Series(tentative_df.index)
                    name_series.to_csv(
                        os.path.join(
                            dirname, "reaction_names_" + name + "_rounded.csv"
                        ),
                        header=False,
                        index=False,
                    )
                else:
                    tentative_df.to_csv(
                        os.path.join(dirname, attribute + "_" + name + "_rounded.csv"),
                        header=False,
                        index=False,
                    )


class HDF5:
    @staticmethod
    def polytope_to_h5(polytope, filename):
        """
        Writes a Polytope object to an HDF5 file
        :param polytope:
        :param filename:
        :return:
        """

        # hf.create_dataset('start', data=x)

        if os.path.exists(filename):
            os.remove(filename)
        # hf = h5py.File(filename, 'w')
        for attribute in dir(polytope):
            tentative_df = getattr(polytope, attribute)
            if isinstance(tentative_df, pd.DataFrame) or isinstance(
                tentative_df, pd.Series
            ):
                tentative_df.to_hdf(filename, attribute, mode="a")
                # hf.create_dataset(attribute, data=tentative_df)
        #
        # hf.close()

    @staticmethod
    def h5_to_polytope(filename):
        """
        Reads a Polytope object from an HDF5 file. The HDF5 file should have the attributes:
        A (DataFrame), b (Series) Ax < b
        and optionally
        S (DataFrame), h (Series) Sx = h
        :param filename:
        :return: Polytope object
        """
        A = pd.read_hdf(filename, key="A")
        b = pd.read_hdf(filename, key="b")
        try:
            S = pd.read_hdf(filename, key="S")
            h = pd.read_hdf(filename, key="h")
        except KeyError:
            return Polytope(A, b)
        return Polytope(A, b, S=S, h=h)



import unittest
import sys
import os

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))
# print(sys.path[0])
sys.path.insert(0, os.getcwd())
from PolyRound.static_classes.lp_utils import ChebyshevFinder
from PolyRound.static_classes.hdf5_csv_io import HDF5
import numpy as np
from PolyRound.static_classes.lp_interfacing import OptlangInterfacer
from PolyRound.static_classes.constraint_removal_reduction import PolytopeReducer
from PolyRound.static_classes.rounding.maximum_volume_ellipsoid import (
    MaximumVolumeEllipsoidFinder,
)
from PolyRound.static_classes.rounding.nearest_symmetric_positive_definite import NSPD
from PolyRound.static_classes.rounding.geometric_mean_scaling import (
    geometric_mean_scaling,
)
from PolyRound.mutable_classes.polytope import Polytope
from PolyRound.static_classes.parse_sbml_stoichiometry import StoichiometryParser
from PolyRound.api import PolyRoundApi
import pandas as pd


class TestHDF5IO(unittest.TestCase):
    def test_reading(self):
        polytope1 = StoichiometryParser.parse_sbml_cobrapy(
            "PolyRound/models/e_coli_core.xml"
        )
        filename = "PolyRound/output/temp.hdf5"
        HDF5.polytope_to_h5(polytope1, filename)
        polytope2 = HDF5.h5_to_polytope(filename)
        for attribute in dir(polytope1):
            tentative_df1 = getattr(polytope1, attribute)
            tentative_df2 = getattr(polytope2, attribute)
            if isinstance(tentative_df1, pd.DataFrame) or isinstance(
                tentative_df1, pd.Series
            ):
                test = tentative_df1.equals(tentative_df2)
                self.assertTrue(test)


class TestRounding(unittest.TestCase):
    def test_solve(self):
        A = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, -1], [1, 1, 1]])
        b = np.array([[0], [0], [0], [1]])
        x = np.array([[0.1], [0.1], [0.1]])

        x, E, converged = MaximumVolumeEllipsoidFinder.run_mve(A, b, x, 1e-3)
        E_sol = np.array(
            [[0.25, 0, 0], [-0.0833, 0.2357, 0], [-0.0833, -0.1178, 0.2042]]
        )
        self.assertTrue(np.allclose(E, E_sol, atol=1e-03))

    def test_solve_coli(self):
        A = np.loadtxt("PolyRound/csvs/A_ineq_e_coli_core_test.csv", delimiter=",")
        b = np.loadtxt("PolyRound/csvs/b_ineq_e_coli_core_test.csv", delimiter=",")[
            :, None
        ]
        E_sol = np.loadtxt("PolyRound/csvs/e_coli_core_mve_test_sol.csv", delimiter=",")
        x = np.zeros((A.shape[1], 1))
        x, E, converged = MaximumVolumeEllipsoidFinder.run_mve(A, b, x, 1e-3)
        diff_mat = np.abs(E_sol - E)
        self.assertLess(np.min(diff_mat), 1e-10)
        pass

    def test_iterative_solve(self):
        A = np.loadtxt("PolyRound/csvs/A_ineq_e_coli_core_test.csv", delimiter=",")
        b = np.loadtxt("PolyRound/csvs/b_ineq_e_coli_core_test.csv", delimiter=",")[
            :, None
        ]
        A_sol = np.loadtxt(
            "PolyRound/csvs/A_e_coli_core_mve_test_reg.csv", delimiter=","
        )
        b_sol = np.loadtxt(
            "PolyRound/csvs/b_e_coli_core_mve_test_reg.csv", delimiter=","
        )
        p = Polytope(A, np.squeeze(b))
        MaximumVolumeEllipsoidFinder.iterative_solve(p, backend="glpk")
        diff_mat = np.abs(p.A.values - A_sol)
        self.assertLess(np.min(diff_mat), 1e-10)
        diff_vec = np.abs(p.b.values - b_sol)
        self.assertLess(np.min(diff_vec), 1e-10)

    def test_geometric_mean_scaling(self):
        A = np.array([[1, 2, 3, 4], [1e1, 1e2, 1e3, 1e4], [1e-1, 1e-2, 1e-3, 1e-4]])
        cscale, rscale = geometric_mean_scaling(A, 0, 0.99)
        cscale_sol = np.array(
            [31.6227766016838, 3.16227766016838, 3.16227766016838, 31.6227766016838]
        )
        rscale_sol = np.array([2.00000000000000, 316.227766016837, 0.00316227766016838])
        r_diff = np.max(rscale - rscale_sol)
        c_diff = np.max(cscale - cscale_sol)
        self.assertLess(r_diff, 1e-10)
        self.assertLess(c_diff, 1e-10)

    def test_NSPD(self):
        A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        A_sol = np.array([[2, 3, 4], [3, 5, 6], [4, 6, 9]])
        A_hat = NSPD.get_NSPD(A)
        diff = np.max(np.abs(A_hat - A_sol))
        self.assertLess(diff, 1)

    def test_transform(self):
        S = np.zeros((1, 3))
        S[0, :] = 1
        A = pd.DataFrame(np.vstack((np.eye(3), -np.eye(3), S)))
        b = pd.Series(np.array([1, 1, 1, 1, 1, 1, 0], dtype=float))
        x = np.array([4, 9, 83])
        polytope = Polytope(A, b, S=S)
        shift1 = np.array([1, 2, 3])
        transform1 = np.array([[43, 6, 3], [11, 6, 4], [-54, 5, 5431]])
        shift2 = np.array([-5, 3, 0.41])
        transform2 = np.matmul(shift1[:, None], shift2[:, None].T) + np.eye(3)
        x_mod = x.copy()
        x_mod -= shift1
        polytope.apply_shift(shift1)
        x_mod = np.matmul(np.linalg.inv(transform1), x_mod)
        polytope.apply_transformation(transform1)
        self.assertTrue(np.allclose(x, polytope.back_transform(x_mod.copy())))
        x_mod -= shift2
        polytope.apply_shift(shift2)
        self.assertTrue(np.allclose(x, polytope.back_transform(x_mod.copy())))
        x_mod = np.matmul(np.linalg.inv(transform2), x_mod)
        polytope.apply_transformation(transform2)
        self.assertTrue(np.allclose(x, polytope.back_transform(x_mod.copy())))
        shift3 = np.power(shift1, shift2)
        x_mod -= shift3
        polytope.apply_shift(shift3)
        self.assertTrue(np.allclose(x, polytope.back_transform(x_mod.copy())))


class TestCoordinateTransform(unittest.TestCase):
    def setUp(self):
        self.input = "PolyRound/models/e_coli_core.xml"
        self.S_cobra = np.loadtxt("PolyRound/csvs/S_cobra.csv", delimiter=",")
        self.A_cobra = np.loadtxt("PolyRound/csvs/A_cobra.csv", delimiter=",")
        self.b_cobra = np.loadtxt("PolyRound/csvs/b_cobra.csv", delimiter=",")
        # self.options = {"presolve": True, "method": "revised simplex"}

    def test_degenerate_polytope(self):
        S = np.array([[1, 0], [0, 1],])
        A = np.eye(2)
        b = np.ones(2)
        p = Polytope(A, b, S=S)
        self.assertRaises(ValueError, PolyRoundApi.simplify_polytope, p)

    def test_simple_null_space(self):
        S = np.array([[1, -1, 0], [0, 1, -1],])
        null = PolytopeReducer.null_space(S)
        self.assertEqual(null.shape[1], 1)

    def test_keep_equalities(self):
        S = np.array([[1, -1, 0], [0, 1, -1],])
        A = np.array([[0.0, 0.0, 1.0], [0.0, 0.0, -1.0]])
        b = np.array([5.25, 1.75])
        p = Polytope(A, b, S=S)
        p = PolyRoundApi.simplify_transform_and_round(p)
        self.assertTrue(p.transformation.shape == (3, 1))

    def test_interior_point_original(self):
        backend = "glpk"
        polytope = StoichiometryParser.parse_sbml_cobrapy(self.input)

        obj = np.ones(polytope.A.shape[1])
        b_eq = np.zeros(polytope.S.shape[0])
        S = np.array(polytope.S).astype(float)

        A = np.array(polytope.A).astype(float)

        b = np.array(polytope.b).astype(float)
        sol, model = OptlangInterfacer.gurobi_solve(obj, A, b, backend, S=S, h=b_eq)
        self.assertEqual(model.status, "optimal")

    def test_matlab_matrices(self):
        backend = "glpk"
        obj = np.ones(self.S_cobra.shape[1])
        b_eq = np.zeros(self.S_cobra.shape[0])
        sol, model = OptlangInterfacer.gurobi_solve(
            obj, self.A_cobra, self.b_cobra, backend, S=self.S_cobra, h=b_eq
        )
        self.assertEqual(model.status, "optimal")

    def test_compare_python_cobra_matrices(self):

        polytope = StoichiometryParser.parse_sbml_cobrapy(self.input)

        S_cobra = self.S_cobra / 10000

        for col, cob_col in zip(np.transpose(polytope.S.values), np.transpose(S_cobra)):
            sim = col == cob_col
            self.assertTrue(sim.all())

        A_cobra = np.around(self.A_cobra)
        A_cobra = A_cobra.astype(int)
        A = np.array(polytope.A, dtype=int)
        for col, cob_col in zip(np.transpose(A), np.transpose(A_cobra)):
            sim = col == cob_col
            self.assertTrue(sim.all())

        # Now move on to b
        b = np.array(polytope.b)
        b = b.reshape((b.size,))
        bool = self.b_cobra == b
        self.assertTrue(bool.all())

    def test_chebyshev_center(self):
        S = np.zeros((1, 3))
        S[0, :] = 1

        A = pd.DataFrame(np.vstack((np.eye(3), -np.eye(3), S)))
        b = pd.Series(np.array([1, 1, 1, 1, 1, 1, 0], dtype=float))
        x, dist0 = ChebyshevFinder.chebyshev_center(Polytope(A, b))
        self.assertTrue(np.all(np.abs(x - x[0]) < 1e-10))
        self.assertTrue(dist0 > 0)

        b[3] = 0
        x, dist1 = ChebyshevFinder.chebyshev_center(Polytope(A, b))
        self.assertTrue(np.all(x[0] - x[1:] > 0))
        self.assertTrue(dist0 > dist1)

    def test_minimal_lp(self):
        b = np.array([1, 1, 1, 1])
        A_ext = np.array([[1, 0, 1], [0, 1, 1], [-1, 0, 1], [0, -1, 1]])
        obj = np.array([0, 0, -1])
        val, m = OptlangInterfacer.gurobi_solve(obj, A_ext, b, "glpk")
        self.assertTrue(np.all(val == np.array([0, 0, 1])))
        # p = Polytope(A, b)
        # x, dist = ChebyshevFinder.chebyshev_center(p)
        # self.assertTrue(False)

    def test_constraint_removal(self):
        S = np.zeros((1, 3))
        S[0, :] = [0, 0, 1]
        h = np.zeros((S.shape[0],))
        h[0] = 1
        A = np.vstack((-np.eye(3), -np.ones((1, 3)), np.ones((1, 3))))
        b = np.array([1, 1, 1, -1, 1], dtype=float)
        b = b.reshape((5,))
        p = Polytope(pd.DataFrame(A), pd.Series(b), S=pd.DataFrame(S), h=pd.Series(h))
        reduced_polytope, removed, refunctioned = PolytopeReducer.constraint_removal(
            p, dict()
        )
        A_true = np.array([[-1, 0, 0], [0, -1, 0]])
        S_true = np.array([[-1, -1, -1], [0, 0, 1]])
        b_true = [1, 1]
        h_true = [-1, 1]
        self.assertTrue(np.all(reduced_polytope.A == A_true))
        self.assertTrue(np.all(reduced_polytope.S == S_true))
        self.assertTrue(np.all(reduced_polytope.b == b_true))
        self.assertTrue(np.all(reduced_polytope.h == h_true))

    def test_constraint_simplification_non_bounded(self):
        A = np.vstack((-np.eye(3), -np.ones((1, 3))))
        b = np.array([1, 1, 1, 3], dtype=float)
        b = b.reshape((4,))
        p = Polytope(pd.DataFrame(A), pd.Series(b))
        reduced_polytope = PolyRoundApi.simplify_polytope(p)
        A_true = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, -1]])
        b_true = [1, 1, 1]
        self.assertTrue(np.all(reduced_polytope.A == A_true))

        self.assertTrue(np.all(reduced_polytope.b == b_true))

    def test_transformation_coli(self):
        backend = "gurobi"
        # eps is for svd
        eps = 1e-12
        # thresh is for declaring equality constraints
        thresh = 1e-06

        polytope = StoichiometryParser.parse_sbml_cobrapy(self.input)
        polytope.normalize()
        reduced_polytope = PolyRoundApi.simplify_polytope(polytope)

        # compute chebyshev center
        x, dist = ChebyshevFinder.chebyshev_center(
            reduced_polytope,
            backend=backend,
            regularize=backend == "gurobi",
            NumericFocus=3,
            MarkowitzTol=0.999,
            FeasibilityTol=1e-09,
            OutputFlag=0,
        )
        self.assertGreater(dist, 0.1)
        # print("chebyshev distance is : " + str(dist))

        pre_b_dist = reduced_polytope.border_distance(x)
        self.assertGreater(pre_b_dist, 0.1)
        # print("border distance pre-transformation is: " + str(pre_b_dist))

        # put x at zero!
        reduced_polytope.apply_shift(x)
        # b = b - np.squeeze(np.matmul(A, x))
        x_0 = np.zeros(x.shape)
        b_dist_at_zero = reduced_polytope.border_distance(x_0)
        self.assertGreater(b_dist_at_zero, 0.1)
        self.assertAlmostEqual(pre_b_dist, b_dist_at_zero)
        # print("border distance zero-transformation is: " + str(b_dist_at_zero))

        transformation = PolytopeReducer.null_space(reduced_polytope.S.values, eps=eps)
        reduced_polytope.apply_transformation(transformation)
        # A_null = np.matmul(A, null)
        u = np.zeros((transformation.shape[1], 1))

        norm_check = np.linalg.norm(np.matmul(polytope.S.values, transformation))
        self.assertLess(norm_check, 1e-10)
        # print("norm of the null space is: " + str(norm_check))

        b_dist = reduced_polytope.border_distance(u)
        self.assertAlmostEqual(b_dist, pre_b_dist)

        # test if we can reproduce the original x
        x_rec = reduced_polytope.back_transform(u)

        self.assertLess(np.min(np.abs(x - x_rec)), 1e-10)

        # Do FVA and see if the points are feasible in the original space
        points = ChebyshevFinder.fva(reduced_polytope, backend="glpk")

        # convert all points to original space
        back = reduced_polytope.back_transform(points.values)
        b_dists = polytope.border_distance(back)
        self.assertGreater(b_dists, -1e-9)
        # print("done")

    def test_api_simplification_coli(self):
        polytope = StoichiometryParser.parse_sbml_cobrapy(self.input)
        x, dist = ChebyshevFinder.chebyshev_center(polytope)
        self.assertTrue(np.abs(dist) < 1e-10)
        reduced = PolyRoundApi.simplify_polytope(polytope)
        x, dist = ChebyshevFinder.chebyshev_center(reduced)
        self.assertTrue(np.abs(dist) > 1e-10)
        transformed = PolyRoundApi.transform_polytope(reduced, verbose=True)
        x, dist = ChebyshevFinder.chebyshev_center(transformed)
        self.assertTrue(np.abs(dist) > 1e-10)
        rounded = PolyRoundApi.round_polytope(transformed)
        self.assertTrue(rounded.inequality_only)
        x, dist = ChebyshevFinder.chebyshev_center(rounded)
        self.assertTrue(np.abs(dist) > 1e-10)
        PolyRoundApi.polytope_to_csvs(
            rounded, os.path.join("PolyRound", "output", "export_test")
        )


if __name__ == "__main__":
    unittest.main()



import pandas as pd
from PolyRound.mutable_classes.polytope import Polytope
import cobra


class StoichiometryParser:
    @staticmethod
    def parse_sbml_cobrapy(file, inf_bound=1e5):
        model = StoichiometryParser.read_sbml_model(file)
        p = StoichiometryParser.extract_polytope(model, inf_bound=inf_bound)
        return p

    @staticmethod
    def read_sbml_model(file):
        model = cobra.io.read_sbml_model(file)
        return model

    @staticmethod
    def extract_polytope(model, inf_bound=1e5):
        S = cobra.util.array.create_stoichiometric_matrix(model, array_type="DataFrame")
        # make bounds matrix
        n_react = len(model.reactions)
        A = pd.DataFrame(0.0, index=range(n_react * 2), columns=S.columns)
        b = pd.Series(0.0, index=range(n_react * 2))
        for ind, r in enumerate(list(model.reactions)):
            if r.bounds[1] == float("inf"):
                b[ind] = inf_bound
            else:
                b[ind] = r.bounds[1]
            if r.bounds[0] == float("-inf"):
                b[ind + n_react] = inf_bound
            else:
                b[ind + n_react] = -r.bounds[0]
            A.loc[ind, r.id] += 1
            A.loc[ind + n_react, r.id] -= 1
        p = Polytope(A, b, S=S)
        return p



import numpy as np
import pandas as pd
import copy


class Polytope:
    """
    Polytope class which holds a set of inequality constraints:
    Ax<=b
    and equality constraints:
    Sx=h
    A and S are Pandas Dataframes and b and h are Pandas Series. Column names of A and S are the reaction names
    (preserved). The polytope can be transformed with apply_shift and apply_transformation. Each of these
    transformations applied are stored in the attributes shift and transformation, allowing to back transform points x
    to the original space using the method back_transform.
    """

    def __init__(self, A, b, S=None, h=None):
        self.A = pd.DataFrame(A)
        self.b = pd.Series(b)
        self.A, self.b = Polytope.remove_zero_rows(self.A, self.b)
        self.shift = pd.Series(0, index=self.A.columns)
        self.transformation = pd.DataFrame(
            np.eye(len(self.A.columns)), index=self.A.columns
        )
        # Differentiate the case with equality constraints
        if S is None:
            if h is not None:
                # You can never have a RHS without a LHS
                raise ValueError
            self.inequality_only = True
            self.S = None
            self.h = None
        else:
            self.inequality_only = False
            self.S = pd.DataFrame(S)
            if h is None:
                self.h = pd.Series(np.zeros((S.shape[0])), index=self.S.index)
            else:
                self.h = pd.Series(h)
            self.S, self.h = Polytope.remove_zero_rows(self.S, self.h)

    def border_distance(self, x):
        """
        Computes shortest distance from x to polytope border
        :param x: Numpy array
        :return: Float
        """
        temp = (self.b.values - np.squeeze(np.matmul(self.A.values, x).T)).T
        return np.min(temp)

    def apply_shift(self, shift):
        """
        Shifts Polytope region without altering the shape
        :param shift: Numpy Array
        :return:
        """
        self.shift += np.matmul(
            self.transformation.values, shift.reshape((shift.size,))
        )
        self.b -= np.squeeze(np.matmul(self.A.values, shift))
        # self.center -= shift
        if not self.inequality_only:
            self.h -= np.squeeze(np.matmul(self.S.values, shift))

    def apply_transformation(self, transformation):
        """
        Applies transformation matrix to Polytope. If non-symmetric transformation, assumes reduction to system of only
        inequlity constraints (not checked).
        :param transformation: Numpy Array or Pandas Dataframe
        :return:
        """
        self.transformation = self.transformation.dot(transformation)
        self.A = self.A.dot(transformation)

        # if applying non-symmetric transform, reduce system to only inequality constraints
        if transformation.shape[0] != transformation.shape[1]:
            self.inequality_only = True
            self.S = None
            self.h = None

    def back_transform(self, x):
        """
        Backtransforms x to the original space of the Polytope.
        :param x: Numpy array
        :return:
        """
        return (
            np.squeeze(np.matmul(self.transformation.values, x)).T + self.shift.values
        ).T

    def normalize(self):
        """
        Normalizes each row sum of self.A to 1. Does not change feasible space.
        :return:
        """
        self.A, self.b = Polytope.normalize_system(self.A, self.b)
        if not self.inequality_only:
            self.S, self.h = Polytope.normalize_system(self.S, self.h)

    def copy(self):
        """
        Deep copy of Polytope.
        :return: Polytope
        @rtype: object
        """
        return copy.deepcopy(self)

    @staticmethod
    def remove_zero_rows(S, h):
        row_norm = np.linalg.norm(S, axis=1)
        S = S.iloc[row_norm != 0, :]
        assert (h.iloc[row_norm == 0] == 0).all()
        h = h.iloc[row_norm != 0]
        return S, h

    @staticmethod
    def normalize_system(A, b, row_norm=None):
        if row_norm is None:
            row_norm = np.linalg.norm(A, axis=1)
        row_norm[row_norm == 0] = 1
        A = (A.T / row_norm).T
        b = (b.T / row_norm).T
        return A, b



import numpy as np

import scipy.sparse as sp
from PolyRound.static_classes.lp_interfacing import OptlangInterfacer
from PolyRound.mutable_classes.polytope import Polytope
from PolyRound.static_classes.lp_utils import ChebyshevFinder
import pandas as pd
from sympy.core import Add, Mul
from cobra.util.solver import solvers
import time


class PolytopeReducer:
    @staticmethod
    def constraint_removal(
        polytope, hp_flags, backend="glpk", thresh=1e-7, verbose=False, sgp=False
    ):
        """
        Removes redundant constraints and removes narrow directions by turning them into equality constraints
        :param polytope: Polytope object to round
        :param hp_flags: Dictionary of gurobi flags for high precision solution
        :param thresh: Float determining how narrow a direction has to be to declare an equality constraint
        :param verbose: Bool regulating output level
        :return: Polytope object with non-empty interior and no redundant constraints, number of removed constraints,
        number of inequality constraints turned to equality constraints.
        """
        interface = solvers[backend]
        m = interface.Model()
        m.configuration.timeout = 60 * 5
        # m = optModel("model")
        if backend == "gurobi":
            OptlangInterfacer.configure_gurobi_model(
                m, hp_flags, sgp=sgp, verbose=verbose
            )
        else:
            OptlangInterfacer.configure_optlang_model(m, hp_flags)
        r_names = [str(r) for r in polytope.A.columns]
        x = np.array([interface.Variable(name, lb=None) for name in r_names])
        m.add(x)

        # make all inequality constraints
        OptlangInterfacer.make_sparse_constraint_system(
            m, polytope.A, polytope.b, x, backend
        )
        # m.addConstr(A @ x <= np.squeeze(polytope.b.values), name="constr")

        if polytope.S is not None:
            # S = sp.csr_matrix(polytope.S)
            # m.addConstr(S @ x == np.squeeze(polytope.h.values), name="eq")
            OptlangInterfacer.make_sparse_constraint_system(
                m, polytope.S, polytope.h, x, backend, equality=True
            )

        m.update()
        # get Chebyshev center for faster solutions
        # cheb_center, dist = ChebyshevFinder.chebyshev_center(
        #     polytope, sgp=sgp, **hp_flags
        # )
        # cheb_center = pd.Series(cheb_center.squeeze(), index=r_names)
        # start = time.time()
        m, removed, refunctioned = PolytopeReducer.constraint_removal_loop(
            m, thresh, backend, verbose=verbose
        )
        # print(time.time() - start)
        A, b = OptlangInterfacer.constraints_as_mat(m, r_names, sense="<")
        if polytope.S is not None:
            S, h = OptlangInterfacer.constraints_as_mat(m, r_names, sense="=")
        if verbose:
            print("Number of removed constraints: " + str(removed))
            print("Number of refunctioned constraints: " + str(refunctioned))
        if polytope.S is not None:
            reduced_polytope = Polytope(A, b, S, h)
            return reduced_polytope, removed, refunctioned
        else:
            reduced_polytope = Polytope(A, b)
            return reduced_polytope, removed, refunctioned

    @staticmethod
    def constraint_removal_loop(m, thresh, backend, start_ind=None, verbose=False):
        interface = solvers[backend]
        constrs = m.constraints
        constr_index = {i: c for i, c in enumerate(constrs) if c.lb != c.ub}
        # in case all constraints are inequalities, we only try remove redundancies
        no_eq_constraints = len(constrs) == len(constr_index)
        if start_ind is not None:
            constr_index = {i: c for i, c in constr_index.items() if i >= start_ind}
        # for each constraint, solve three lps
        removed = 0
        refunctioned = 0
        # avoided_lps = 0
        for i in constr_index:
            if verbose:
                if i % 50 == 0:
                    print("\n Investigating constraint number: " + str(i) + "\n")

                # m.write("output/debug_temp.mps")
            constr_expr = constr_index[i].expression
            m.objective = interface.Objective(constr_expr, direction="max")
            # m.setObjective(constr_expr, gp.GRB.MAXIMIZE)
            # m.update()
            m.optimize()
            max_val = OptlangInterfacer.get_opt(m, backend)

            # now get altered problem
            orig_rhs = constr_index[i].ub
            constr_index[i].ub = orig_rhs + 1
            # m.update()
            m.optimize()
            pert_val = OptlangInterfacer.get_opt(m, backend)
            constr_index[i].ub = orig_rhs
            if np.abs(max_val - pert_val) < thresh:
                removed += 1
                # In this case remove constraint
                m.remove(constr_index[i])
            elif not no_eq_constraints:
                # Check if it might be an equality
                # first, make the "cheap" check
                # chebVal = PolytopeReducer.linExprSeriesProd(cheb_center, constr_expr)
                # if np.abs(max_val - chebVal) > thresh:
                #     avoided_lps += 1
                #     continue

                # m.setObjective(constr_expr, gp.GRB.MINIMIZE)
                m.objective = interface.Objective(constr_expr, direction="min")
                # m.update()
                m.optimize()
                min_val = OptlangInterfacer.get_opt(m, backend)
                if np.abs(max_val - min_val) < thresh:
                    refunctioned += 1
                    constr_index[i].lb = constr_index[i].ub
                    # center = (max_val + min_val) / 2
                    # if center != constr_index[i].RHS:
                    #     constr_index[i].setAttr("rhs", center)

            m.update()
        # if verbose:
        #     print(str(avoided_lps) + " lps were avoided with chebyshev center speed up")
        return m, removed, refunctioned

    @staticmethod
    def null_space(S, eps=1e-10):
        """
        Returns the null space of a matrix
        :param S: Numpy array
        :param eps: Threshold for declaring 0 singular values
        :return: Numpy array of null space
        """
        u, s, vh = np.linalg.svd(S)
        s = np.array(s.tolist())
        vh = np.array(vh.tolist())
        null_mask = s <= eps
        null_mask = np.append(null_mask, True)
        null_ind = np.argmax(null_mask)
        null = vh[null_ind:, :]
        return np.transpose(null)

    @staticmethod
    def linExprSeriesProd(center, lin_expr):
        val = 0.0
        if isinstance(lin_expr, Add):
            for mul in lin_expr.args:
                assert isinstance(mul, Mul)
                val += PolytopeReducer.multiplyMulObject(center, mul)
        else:
            assert isinstance(lin_expr, Mul)
            val += PolytopeReducer.multiplyMulObject(center, lin_expr)
        return val

    @staticmethod
    def multiplyMulObject(center, mul):
        val = 0.0
        mul_dict = mul.as_coefficients_dict()
        for name, coeff in mul_dict.items():
            val += center[name.name] * coeff
        return val



import numpy as np
import pandas as pd
from PolyRound.static_classes.lp_interfacing import OptlangInterfacer


class ChebyshevFinder:
    @staticmethod
    def chebyshev_center(
        polytope, backend="glpk", regularize=False, sgp=False, **kwargs
    ):
        # get norm col
        a_norm = np.linalg.norm(polytope.A.values, axis=1).reshape(
            (polytope.A.shape[0], 1)
        )
        A_ext = np.concatenate((polytope.A.values, a_norm), axis=1)
        obj = np.zeros(A_ext.shape[1])
        obj[-1] = -1
        if polytope.inequality_only:
            x, m = OptlangInterfacer.gurobi_solve(
                obj, A_ext, polytope.b.values, backend, sgp=sgp, **kwargs
            )
        else:
            s_0_col = np.zeros(shape=(polytope.S.shape[0], 1))
            S_ext = np.concatenate((polytope.S.values, s_0_col), axis=1)
            # sol = solvers.lp(matrix(obj), matrix(A_ext), matrix(b_np), A=matrix(S_ext), b=matrix(h), solver='glpk')
            # print(sol)
            x, m = OptlangInterfacer.gurobi_solve(
                obj,
                A_ext,
                polytope.b.values,
                backend,
                S=S_ext,
                h=polytope.h.values,
                sgp=sgp,
                **kwargs
            )
        if regularize:
            x_reg, m = OptlangInterfacer.gurobi_regularize_chebyshev_center(
                x[-1], m, backend
            )
            x = x_reg
        x = x.reshape((x.shape[0], 1))
        return x[:-1], x[-1]

    @staticmethod
    def fva(polytope, backend):
        n_reac = polytope.A.shape[1]
        output = pd.DataFrame(index=polytope.A.columns)
        # make the first run
        obj = np.ones(n_reac)

        if polytope.inequality_only:
            x, m = OptlangInterfacer.gurobi_solve(
                obj, polytope.A.values, polytope.b.values, backend
            )
        else:
            x, m = OptlangInterfacer.gurobi_solve(
                obj,
                polytope.A.values,
                polytope.b.values,
                backend,
                S=polytope.S.values,
                h=polytope.h.values,
            )

        obj = np.zeros(n_reac)
        # Now run all the remaining directions
        for i in range(0, n_reac * 2):
            ind = i // 2
            if i % 2 == 0:
                obj[ind] = 1
            else:
                obj[ind] = -1
            x, m = OptlangInterfacer.gurobi_solve_model(obj, m, backend)
            obj[ind] = 0
            output.loc[:, i] = x
        return output



import numpy as np


def geometric_mean_scaling(A, iprint, scltol):
    if iprint > 0:
        print("gmscale: Geometric-Mean scaling of matrix")
        print("ax col ratio")

    [m, n] = A.shape
    A = np.abs(A)
    maxpass = 150
    aratio = 1e50
    damp = 1e-4
    rscale = np.ones((m, 1))
    cscale = np.ones((n, 1))

    for npass in range(maxpass + 1):

        # Find the largest column ratio.
        # Also set new column scales (except on pass 0).

        rscale[rscale == 0] = 1
        # Rinv    = diag(sparse(1./rscale));
        SA = np.multiply(A.transpose(), 1 / rscale.squeeze()).transpose()  # Rinv*A
        # [I,J,V] = find(SA);
        with np.errstate(divide="ignore"):
            invSA = 1 / SA  # sparse(I,J,1./V,m,n);
        cmax = np.max(SA, axis=0).transpose()  # full(max(   SA))';   % column vector
        cmin = np.max(invSA, axis=0)  # full(max(invSA))';   % column vector
        cmin = 1.0 / (cmin + np.spacing(1))
        with np.errstate(divide="ignore"):
            sratio = np.max(cmax / cmin)
        if npass > 0:
            cscale = np.sqrt(np.maximum(cmin, damp * cmax) * cmax)

        if iprint > 0:
            print(npass, sratio)

        if npass >= 2 and sratio >= aratio * scltol:
            break
        if npass == maxpass:
            break
        aratio = sratio

        # Set new row scales for the next pass.

        cscale[cscale == 0] = 1
        # Cinv    = diag(sparse(1./cscale));
        SA = np.multiply(A, 1 / cscale.squeeze())  # A*Cinv;                  % Scaled A
        # [I,J,V] = find(SA);
        with np.errstate(divide="ignore"):
            invSA = 1 / SA  # sparse(I,J,1./V,m,n);
        rmax = np.max(SA, axis=1).transpose()  # full(max(   SA))';   % column vector
        rmin = np.max(invSA, axis=1)  # full(max(invSA))';   % column vector
        rmin = 1.0 / (rmin + np.spacing(1))
        rscale = np.sqrt(np.maximum(rmin, damp * rmax) * rmax)

    # Reset column scales so the biggest element
    # in each scaled column will be 1.
    # Again, allow for empty rows and columns.

    rscale[rscale == 0] = 1
    # Rinv = diag(sparse(1. / rscale));
    SA = np.multiply(A.transpose(), 1 / rscale).transpose()
    # [I, J, V] = find(SA);
    cscale = np.max(SA, axis=0).transpose()
    cscale[cscale == 0] = 1

    return cscale, rscale



import numpy as np
from PolyRound.static_classes.rounding.geometric_mean_scaling import (
    geometric_mean_scaling,
)
from PolyRound.static_classes.rounding.nearest_symmetric_positive_definite import NSPD
from PolyRound.static_classes.lp_utils import ChebyshevFinder
from scipy.linalg import cho_factor, cho_solve
from PolyRound.default_settings import (
    default_0_width,
    default_hp_flags,
    default_max_ratio_bmAx0,
)


class MaximumVolumeEllipsoidFinder:
    @staticmethod
    def iterative_solve(
        polytope, backend, hp_flags=default_hp_flags, verbose=False, sgp=False
    ):
        # only round reduced polytopes
        assert polytope.inequality_only
        nDimensions = polytope.A.shape[1]
        columnScale, rowScale = geometric_mean_scaling(polytope.A.values, 0, 0.99)
        # just this once, do the wasteful expansion
        c_scale_transform = np.diag(1 / columnScale)
        polytope.apply_transformation(c_scale_transform)

        # Normalize
        polytope.normalize()

        # Perform iterative rounding
        maxIterations = 20
        iteration = 0
        reg = 1e-3
        iterationTransform = np.eye(nDimensions)
        converged = 0
        temp_eig = np.linalg.eig(iterationTransform)[0]
        while (
            (np.max(temp_eig) > 6 * np.min(temp_eig) and converged != 1)
            or reg > 1e-8
            or converged == 2
        ):
            del temp_eig
            iteration = iteration + 1
            [center, distance] = ChebyshevFinder.chebyshev_center(
                polytope, backend=backend, sgp=sgp, **hp_flags
            )
            reg = np.maximum(reg / 10, 1e-10)

            # Calculate and apply transform
            (
                iterationShift,
                iterationTransform,
                converged,
            ) = MaximumVolumeEllipsoidFinder.run_mve(
                polytope.A.values.copy(),
                polytope.b.values.copy()[:, None],
                center,
                reg,
                verbose=verbose,
            )
            polytope.apply_shift(iterationShift)
            polytope.apply_transformation(iterationTransform)
            polytope.normalize()
            if iteration == maxIterations:
                break
            temp_eig_all = np.linalg.eig(iterationTransform)
            temp_eig = temp_eig_all[0]
            if verbose:
                print(
                    "Iteration: "
                    + str(iteration)
                    + ", reg: "
                    + str(reg)
                    + ", log ellipsoid vol: "
                    + str(np.sum(np.log(temp_eig)))
                    + ", longest axis: "
                    + str(np.max(temp_eig))
                    + ", shortest axis: "
                    + str(np.min(temp_eig))
                    + ", border distance: "
                    + str(distance)
                )

        if iteration == maxIterations:
            if verbose:
                print(
                    "Maximum number of iterations reached, rounding may not be ideal."
                )

        if verbose:
            print("Maximum volume ellipsoid found.")
        if np.min(polytope.b.values) <= 0:
            center, distance = ChebyshevFinder.chebyshev_center(
                polytope, backend=backend, sgp=sgp, **hp_flags
            )
            polytope.apply_shift(center)
            if verbose:
                print(
                    "Shifting so that the origin is inside the polytope. Rounding may not be ideal."
                )

    @staticmethod
    def run_mve(A, b, x0, reg, verbose=False):
        maxiter = 150
        tol2 = 1.0e-6
        x, E2, msg, y, z, iter = MaximumVolumeEllipsoidFinder.solve_mve(
            A, b, x0, reg, maxiter=maxiter, tol=tol2, verbose=verbose
        )
        return x, np.linalg.cholesky(NSPD.get_NSPD(E2)), msg

    @staticmethod
    def check_convergence(
        E2,
        r1,
        r2,
        last_r1,
        last_r2,
        res,
        tol,
        bnrm,
        rmu,
        minmu,
        prev_obj,
        x,
        x0,
        reg,
        iter,
        verbose,
    ):
        msg = 0
        E2_eig_all = np.linalg.eig(E2)
        E2_eig = E2_eig_all[0]
        if np.min(E2_eig) > 0:
            sum_det = np.sum(np.log(E2_eig))
            objval = sum_det / 2
            # objval = np.log(np.linalg.det(E2)) / 2
        else:
            objval = -np.inf

        if verbose:
            print("Iteration number in inner loop of MVE optimization: " + str(iter))
        if (
            np.abs((last_r1 - r1) / np.minimum(np.abs(last_r1), np.abs(r1))) < 1e-2
            and np.abs((last_r2 - r2) / np.minimum(np.abs(last_r2), np.abs(r2))) < 1e-2
            and np.max(E2_eig) / np.min(E2_eig) > 100
            and reg > 1e-10
        ):
            if verbose:
                print("Stopped making progress, stopping and restarting.")
            x = x + x0.squeeze()
            msg = 2

        if (res < tol * (1 + bnrm) and rmu <= minmu) or (
            iter > 100
            and prev_obj != -np.inf
            and (prev_obj >= (1 - tol) * objval or prev_obj <= (1 - tol) * objval)
        ):
            if verbose:
                print("Converged")
            x = x + x0.squeeze()
            msg = 1

        return objval, msg, x

    @staticmethod
    def solve_mve(A, b, x0, reg, maxiter=50, tol=1e-4, verbose=False):
        # check that x and b are 2 dimensional
        assert len(b.shape) == 2 and len(x0.shape) == 2
        m, n = A.shape
        bnrm = np.linalg.norm(b)

        minmu = 1.0e-8
        tau0 = 0.75

        smallest_scaling = default_0_width
        last_r1 = -np.inf
        last_r2 = -np.inf

        bmAx0 = b - np.matmul(A, x0)
        bmAx0 = np.maximum(bmAx0, smallest_scaling)
        min_el = np.min(bmAx0)
        bmAx0 = np.minimum(bmAx0, min_el * default_max_ratio_bmAx0)
        # if np.any(bmAx0 <= 0):
        #     if verbose:
        #         print("x0 not interior, use absolute value")
        #     bmAx0 = np.abs(bmAx0)
        #     # raise ValueError

        A = np.divide(A, bmAx0)
        b = np.ones((m,))
        x = np.zeros((n,))
        y = np.ones((m,))
        bmAx = b

        prev_obj = -np.inf
        astep = None
        Adx = None
        msg = 0
        for iter in range(1, maxiter + 1):

            if astep is not None and Adx is not None:
                bmAx = bmAx - astep * Adx

            Aty = np.multiply(np.transpose(A), np.squeeze(y))
            assert A.shape[0] == Aty.shape[1]
            assert A.shape[1] == Aty.shape[0]
            prod = np.matmul(Aty, A)
            E2 = np.linalg.inv(prod)
            Q = np.matmul(np.matmul(A, E2), np.transpose(A))
            h = np.sqrt(np.diag(Q))
            if iter == 1:
                t = np.min(bmAx / h)
                y = y / np.power(t, 2)
                h = t * h
                z = np.maximum(1.0e-1, np.squeeze(bmAx) - h)
                Q = np.power(t, 2) * Q

            yz = np.squeeze(y) * z
            yh = np.squeeze(y) * h

            gap = np.sum(yz) / m
            rmu = np.minimum(0.5, gap) * gap
            rmu = np.maximum(rmu, minmu)

            R1 = -np.matmul(np.transpose(A), yh)
            R2 = np.squeeze(bmAx) - h - z
            R3 = rmu - yz

            r1 = np.max(np.abs(R1))
            r2 = np.max(np.abs(R2))
            r3 = np.max(np.abs(R3))
            res = np.max([r1, r2, r3])

            if iter % 10 == 0:
                # print(r2, r1, r3, objval);
                objval, msg, x = MaximumVolumeEllipsoidFinder.check_convergence(
                    E2,
                    r1,
                    r2,
                    last_r1,
                    last_r2,
                    res,
                    tol,
                    bnrm,
                    rmu,
                    minmu,
                    prev_obj,
                    x,
                    x0,
                    reg,
                    iter,
                    verbose,
                )
                if msg == 1 or msg == 2:
                    break
                last_r2 = r2
                last_r1 = r1
                prev_obj = objval
            #
            YQ = np.multiply(Q, np.squeeze(y))
            YQQY = YQ * np.transpose(YQ)
            y2h = 2 * yh
            YA = np.transpose(np.multiply(np.transpose(A), np.squeeze(y)))
            G = YQQY
            elementwise_max = np.maximum(reg, y2h * z)
            G.flat[:: G.shape[1] + 1] += elementwise_max
            temp_rhs = np.transpose(np.multiply(np.transpose(YA), h + z))
            G_c, low = cho_factor(G)
            T = cho_solve((G_c, low), temp_rhs)
            temp = np.transpose(np.multiply(np.transpose(T), y2h))
            ATP = np.transpose(temp - YA)
            R3Dy = R3 / np.squeeze(y)
            R23 = R2 - R3Dy
            ATP_A = np.matmul(ATP, A)
            ATP_A.flat[:: ATP_A.shape[1] + 1] += reg
            dx = np.linalg.solve(ATP_A, R1 + np.matmul(ATP, R23))

            Adx = np.matmul(A, dx)
            dyDy = cho_solve((G_c, low), y2h * (Adx - R23))

            dy = y * dyDy
            dz = R3Dy - z * dyDy

            ax = -1 / np.minimum(np.min(-Adx / bmAx), -0.5)
            ay = -1 / np.minimum(np.min(dyDy), -0.5)
            az = -1 / np.minimum(np.min(dz / z), -0.5)
            tau = np.maximum(tau0, 1 - res)
            astep = tau * np.min([1, ax, ay, az])

            x = x + astep * dx
            y = y + astep * dy
            z = z + astep * dz

            if reg > 1e-6 and iter >= 10:
                break

        return x, E2, msg, y, z, iter



import numpy as np


class NSPD:
    @staticmethod
    def get_NSPD(A):
        assert A.shape[0] == A.shape[1]

        B = (A + A.transpose()) / 2
        U, S, V = np.linalg.svd(B)
        US = np.multiply(U, S)
        H = np.matmul(US, U.transpose())
        A_hat = (B + H) / 2
        A_hat = NSPD.correct_numerics(A_hat)
        return A_hat

    @staticmethod
    def correct_numerics(A_hat):
        mineig = 0
        # p = 1
        k = 0
        # while p != 0:
        while mineig <= 0:
            # R, p = np.linalg.cholesky(A_hat)
            eigenvals = np.linalg.eig(A_hat)[0]
            mineig = np.min(eigenvals)
            k += 1
            # if p != 0:
            if mineig <= 0:

                A_hat = A_hat + (
                    -mineig * k ** 2
                    + np.spacing(np.min(np.diag(A_hat))) * np.eye(A_hat.shape[0])
                )
        return A_hat



from setuptools import setup

with open("README.md", "r") as fh:
	long_description = fh.read()

setup(
    name="PolyRound",
    version="0.1.2",
    description="A python package for rounding polytopes.",
    long_description=long_description,
    url="https://gitlab.com/csb.ethz/PolyRound",
    author="Axel Theorell",
    author_email="atheorell@ethz.ch",
    license="MIT License",
    packages=[
        "PolyRound",
        "PolyRound.mutable_classes",
        "PolyRound.static_classes",
        "PolyRound.static_classes.rounding",
    ],
    install_requires=[
        "numpy>=1.18",
        "pandas>=1.0",
        "python-dateutil>=2.8",
        "python-libsbml>=5.18",
        "scipy>=1.4",
        "h5py>=2.10",
        "optlang>=1.4",
        "tables>=3.6",
        "cobra>=0.20",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    python_requires='>=3.6',
)



import numpy as np
import sys
import os

sys.path.insert(0, os.getcwd())
from PolyRound.static_classes.parse_sbml_stoichiometry import StoichiometryParser
from PolyRound.static_classes.hdf5_csv_io import HDF5
import time
import argparse
from PolyRound.api import PolyRoundApi

from PolyRound.default_settings import default_hp_flags
from cobra.util.solver import solvers


if "gurobi" in solvers:
    pref_backend = "gurobi"
else:
    pref_backend = "glpk"


def main(args):
    inputs = args.files
    path = args.path
    thresh = args.thresh
    verbose = args.v
    # set hp flags
    hp_flags = default_hp_flags
    if args.hp:
        hp_flags = {
            "NumericFocus": 3,
            "FeasibilityTol": 1e-09,
            "OptimalityTol": 1e-09,
            "MarkowitzTol": 0.999,
        }

    for input in inputs:
        input_name = input.split("/")[-1].split(".")[0]
        if input.endswith(".xml"):
            polytope = StoichiometryParser.parse_sbml_cobrapy(input)
        else:
            polytope = HDF5.h5_to_polytope(input)
        start_time = time.time()
        if not args.ie:
            polytope = PolyRoundApi.simplify_transform_and_round(
                polytope,
                backend=args.backend,
                hp_flags=hp_flags,
                thresh=thresh,
                verbose=verbose,
                sgp=args.sgp,
            )
        else:
            polytope = PolyRoundApi.round_polytope(polytope, verbose=verbose)
        end_time = time.time()
        HDF5.polytope_to_h5(
            polytope, os.path.join(path, input_name + "_reduced_rounded.hdf5")
        )
        return end_time - start_time


def pars_args():
    # Create the parser
    parser = argparse.ArgumentParser(description="List the content of a folder")

    # Add the arguments
    parser.add_argument(
        "-hp", action="store_true", help="run the program with high precision option"
    )
    parser.add_argument(
        "-ie",
        action="store_true",
        help="input is inequality constraints only (only for hdf5 input)",
    )
    parser.add_argument("-sgp", action="store_true", help="show gurobi progress")
    parser.add_argument("-v", action="store_true", help="run in verbose mode")
    parser.add_argument(
        "-thresh",
        type=float,
        default=1e-7,
        help="Threshold parameter for minimal width of a dimension",
    )
    parser.add_argument(
        "-path", type=str, default="PolyRound/output/", help="Output path"
    )
    parser.add_argument(
        "-backend",
        type=str,
        default=pref_backend,
        help="Backend for solving linear programs. Possible inputs is gurobi and glpk. gurobi is default in case it is available.",
    )
    parser.add_argument("files", nargs="*")

    args = parser.parse_args()
    print(args)
    # Do the input arguments make sense?
    contains_xml = np.any([file.endswith(".xml") for file in args.files])
    if contains_xml and args.ie:
        print("Argument -ie only available for hdf5 inputs")
        raise IOError

    legal_format = np.all(
        [file.endswith(".xml") or file.endswith(".hdf5") for file in args.files]
    )
    if not legal_format:
        print("Only file extensions .xml and .hdf5 allowed")
        raise IOError

    return args


if __name__ == "__main__":
    print(main(pars_args()))



import numpy as np
import pandas as pd
from scipy import sparse as sp
from optlang.symbolics import Zero
from sympy.core import Add, Mul
from cobra.util.solver import solvers
from swiglpk import glp_adv_basis


class OptlangInterfacer:
    @staticmethod
    def make_sparse_constraint_system(m, A, b, x, backend, equality=False):
        interface = solvers[backend]
        A = sp.csr_matrix(A)
        constr_dict = dict()
        for row_ind, bound in enumerate(b):
            if equality:
                constr_dict[interface.Constraint(Zero, ub=bound, lb=bound)] = {
                    x[var_ind]: A[row_ind, var_ind]
                    for var_ind in A[row_ind, :].nonzero()[1]
                }
            else:
                constr_dict[interface.Constraint(Zero, ub=bound)] = {
                    x[var_ind]: A[row_ind, var_ind]
                    for var_ind in A[row_ind, :].nonzero()[1]
                }
        m.add(constr_dict.keys())
        m.update()
        for constr, terms in constr_dict.items():
            constr.set_linear_coefficients(terms)

    @staticmethod
    def configure_gurobi_model(m, hp_flags, sgp=False, verbose=False):
        if not sgp:
            m.problem.setParam("OutputFlag", 0)
        if verbose:
            print("Using the gurobi flags: " + str(hp_flags))
        for key, val in hp_flags.items():
            m.problem.setParam(key, val)
        # Never let the solver multi thread
        m.problem.setParam("Threads", 1)

    @staticmethod
    def configure_optlang_model(m, hp_flags, verbose=False):
        if "FeasibilityTol" in hp_flags:
            m.configuration.tolerances.feasibility = hp_flags["FeasibilityTol"]
        if "OptimalityTol" in hp_flags:
            m.configuration.tolerances.optimality = hp_flags["OptimalityTol"]

    @staticmethod
    def gurobi_solve(obj, A, b, backend, S=None, h=None, sgp=False, **kwargs):
        interface = solvers[backend]
        m = interface.Model()
        if backend == "gurobi":
            OptlangInterfacer.configure_gurobi_model(m, kwargs, sgp=sgp)
        r_names = [str(r) for r in range(A.shape[1])]
        x = np.array([interface.Variable(name, lb=None) for name in r_names])
        m.add(x)
        m.objective = interface.Objective(obj @ x, direction="min")
        # m.setObjective(obj @ x, gp.GRB.MINIMIZE)
        # sp_ratio = np.sum(np.abs(A) > 1e-12) / A.size
        # if sp_ratio < 0.1:
        # A = sp.csr_matrix(A)
        OptlangInterfacer.make_sparse_constraint_system(m, A, b, x, backend)
        # m.addConstr(A @ x <= np.squeeze(b), name="ineq")
        if S is not None:
            assert h is not None
            # sp_ratio = np.sum(np.abs(S) < 1e-12) / S.size
            # if sp_ratio < 0.1:
            # print("Sparse mode")
            # S = sp.csr_matrix(S)
            # m.addConstr(S @ x == np.squeeze(h), name="eq")
            OptlangInterfacer.make_sparse_constraint_system(
                m, S, h, x, backend, equality=True
            )
        # m.update()
        m.optimize()
        if m.status == "optimal":
            return pd.Series(m.primal_values).values, m
        else:
            x = np.zeros(A.shape[1])
            x[:] = np.nan
            return x, m

    @staticmethod
    def gurobi_solve_model(obj, m, backend):
        interface = solvers[backend]
        # m.reset()
        # m.update()
        x = m.variables
        # m.setObjective(obj @ x, gp.GRB.MINIMIZE)
        m.objective = interface.Objective(obj @ x, direction="min")
        # m.update()
        m.optimize()
        if m.status == "optimal":
            return pd.Series(m.primal_values).values, m
        else:
            x = np.zeros(len(m.variables))
            x[:] = np.nan
            return x, m

    @staticmethod
    def gurobi_regularize_chebyshev_center(obj_val, m, backend):
        interface = solvers[backend]
        x = np.array(m.variables)
        last_var = x[-1]
        m.add(interface.Constraint(last_var, lb=obj_val / 2))
        # m.addConstr(last_var, gp.GRB.GREATER_EQUAL, obj_val / 2)
        expr = x @ x
        m.objective = interface.Objective(expr, direction="min")
        # obj = np.eye(len(x))
        # obj[-1, -1] = 0
        # m.setMObjective(obj, None, 0, sense=gp.GRB.MINIMIZE)
        m.update()
        m.optimize()
        return pd.Series(m.primal_values).values, m

    @staticmethod
    def constraints_as_mat(m, r_names, sense="<"):
        constrs = m.constraints
        if sense == "<":
            constrs = [c for c in constrs if c.ub != c.lb]
        elif sense == "=":
            constrs = [c for c in constrs if c.ub == c.lb]
        else:
            raise ValueError
        b = pd.Series([c.ub for c in constrs], dtype=np.float64)
        c_df = pd.DataFrame(
            index=range(len(constrs)), columns=r_names, dtype=np.float64
        )
        for ind, constr in enumerate(constrs):
            if isinstance(constr.expression, Add):
                for mul in constr.expression.args:
                    assert isinstance(mul, Mul)
                    mul_dict = mul.as_coefficients_dict()
                    for var, coeff in mul_dict.items():
                        c_df.loc[ind, var.name] = float(coeff)
            elif isinstance(constr.expression, Mul):
                mul_dict = constr.expression.as_coefficients_dict()
                for var, coeff in mul_dict.items():
                    c_df.loc[ind, var.name] = float(coeff)
            else:
                raise ValueError(
                    "Invalid constraint expression detected, probably caused by a zero row in a constraint matrix."
                )
        c_df.fillna(0.0, inplace=True)
        return c_df, b

    @staticmethod
    def get_opt(m, backend):
        if m.status == "optimal":
            return m.objective.value
        elif m.status == "infeasible" or m.status == "check_original_solver_status":
            print("model in infeasible state, resetting lp")
            if backend == "gurobi":
                m.problem.reset()
            elif backend == "glpk":
                glp_adv_basis(m.problem, 0)
            else:
                raise ValueError(
                    "PolyRound does currently not support system reset for backend "
                    + backend
                )
            m.optimize()
            # m.setParam("BarHomogeneous", -1)
            if m.status == "optimal":
                return m.objective.value
            else:
                raise ValueError
        else:
            print("Solver status: " + str(m.status))
            raise ValueError
